"""Tests for josepy.errors."""
import sys
import unittest

import pytest


class UnrecognizedTypeErrorTest(unittest.TestCase):
    def setUp(self) -> None:
        from josepy.errors import UnrecognizedTypeError

        self.error = UnrecognizedTypeError("foo", {"type": "foo"})

    def test_str(self) -> None:
        assert "foo was not recognized, full message: {'type': 'foo'}" == str(self.error)


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



"""Tests for josepy.b64."""
import sys
from typing import Union

import pytest

# https://en.wikipedia.org/wiki/Base64#Examples
B64_PADDING_EXAMPLES = {
    b"any carnal pleasure.": (b"YW55IGNhcm5hbCBwbGVhc3VyZS4", b"="),
    b"any carnal pleasure": (b"YW55IGNhcm5hbCBwbGVhc3VyZQ", b"=="),
    b"any carnal pleasur": (b"YW55IGNhcm5hbCBwbGVhc3Vy", b""),
    b"any carnal pleasu": (b"YW55IGNhcm5hbCBwbGVhc3U", b"="),
    b"any carnal pleas": (b"YW55IGNhcm5hbCBwbGVhcw", b"=="),
}


B64_URL_UNSAFE_EXAMPLES = {
    bytes((251, 239)): b"--8",
    bytes((255,)) * 2: b"__8",
}


class B64EncodeTest:
    """Tests for josepy.b64.b64encode."""

    @classmethod
    def _call(cls, data: bytes) -> bytes:
        from josepy.b64 import b64encode

        return b64encode(data)

    def test_empty(self) -> None:
        assert self._call(b"") == b""

    def test_unsafe_url(self) -> None:
        for text, b64 in B64_URL_UNSAFE_EXAMPLES.items():
            assert self._call(text) == b64

    def test_different_paddings(self) -> None:
        for text, (b64, _) in B64_PADDING_EXAMPLES.items():
            assert self._call(text) == b64

    def test_unicode_fails_with_type_error(self) -> None:
        with pytest.raises(TypeError):
            # We're purposefully testing with the incorrect type here.
            self._call("some unicode")  # type: ignore


class B64DecodeTest:
    """Tests for josepy.b64.b64decode."""

    @classmethod
    def _call(cls, data: Union[bytes, str]) -> bytes:
        from josepy.b64 import b64decode

        return b64decode(data)

    def test_unsafe_url(self) -> None:
        for text, b64 in B64_URL_UNSAFE_EXAMPLES.items():
            assert self._call(b64) == text

    def test_input_without_padding(self) -> None:
        for text, (b64, _) in B64_PADDING_EXAMPLES.items():
            assert self._call(b64) == text

    def test_input_with_padding(self) -> None:
        for text, (b64, pad) in B64_PADDING_EXAMPLES.items():
            assert self._call(b64 + pad) == text

    def test_unicode_with_ascii(self) -> None:
        assert self._call("YQ") == b"a"

    def test_non_ascii_unicode_fails(self) -> None:
        with pytest.raises(ValueError):
            self._call("\u0105")

    def test_type_error_no_unicode_or_bytes(self) -> None:
        with pytest.raises(TypeError):
            # We're purposefully testing with the incorrect type here.
            self._call(object())  # type: ignore


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



"""Tests for josepy.jwk."""
import binascii
import sys
import unittest
from typing import TYPE_CHECKING

import pytest
import test_util

from josepy import errors, json_util, util

# The approach used here and below is based on
# https://github.com/certbot/certbot/pull/8748.
if TYPE_CHECKING:
    from typing_extensions import Protocol
else:
    Protocol = object

DSA_PEM = test_util.load_vector("dsa512_key.pem")
RSA256_KEY = test_util.load_rsa_private_key("rsa256_key.pem")
RSA512_KEY = test_util.load_rsa_private_key("rsa512_key.pem")
EC_P256_KEY = test_util.load_ec_private_key("ec_p256_key.pem")
EC_P384_KEY = test_util.load_ec_private_key("ec_p384_key.pem")
EC_P521_KEY = test_util.load_ec_private_key("ec_p521_key.pem")


class JWKTest(unittest.TestCase):
    """Tests for josepy.jwk.JWK."""

    def test_load(self) -> None:
        from josepy.jwk import JWK

        with pytest.raises(errors.Error):
            JWK.load(DSA_PEM)

    def test_load_subclass_wrong_type(self) -> None:
        from josepy.jwk import JWKRSA

        with pytest.raises(errors.Error):
            JWKRSA.load(DSA_PEM)


class JWKSubclassTest(Protocol):
    from josepy.jwk import JWK

    jwk: JWK
    thumbprint: bytes


class JWKTestBaseMixin:
    """Mixin test for JWK subclass tests."""

    thumbprint: bytes = NotImplemented

    def test_thumbprint_private(self: JWKSubclassTest) -> None:
        assert self.thumbprint == self.jwk.thumbprint()

    def test_thumbprint_public(self: JWKSubclassTest) -> None:
        assert self.thumbprint == self.jwk.public_key().thumbprint()


class JWKOctTest(unittest.TestCase, JWKTestBaseMixin):
    """Tests for josepy.jwk.JWKOct."""

    thumbprint = (
        b"\xf3\xe7\xbe\xa8`\xd2\xdap\xe9}\x9c\xce>"
        b"\xd0\xfcI\xbe\xcd\x92'\xd4o\x0e\xf41\xea"
        b"\x8e(\x8a\xb2i\x1c"
    )

    def setUp(self) -> None:
        from josepy.jwk import JWKOct

        self.jwk = JWKOct(key=b"foo")
        self.jobj = {"kty": "oct", "k": json_util.encode_b64jose(b"foo")}

    def test_to_partial_json(self) -> None:
        assert self.jwk.to_partial_json() == self.jobj

    def test_from_json(self) -> None:
        from josepy.jwk import JWKOct

        assert self.jwk == JWKOct.from_json(self.jobj)

    def test_from_json_hashable(self) -> None:
        from josepy.jwk import JWKOct

        hash(JWKOct.from_json(self.jobj))

    def test_load(self) -> None:
        from josepy.jwk import JWKOct

        assert self.jwk == JWKOct.load(b"foo")

    def test_public_key(self) -> None:
        assert self.jwk.public_key() is self.jwk


class JWKRSATest(unittest.TestCase, JWKTestBaseMixin):
    """Tests for josepy.jwk.JWKRSA."""

    thumbprint = (
        b"\x83K\xdc#3\x98\xca\x98\xed\xcb\x80\x80<\x0c"
        b"\xf0\x95\xb9H\xb2*l\xbd$\xe5&|O\x91\xd4 \xb0Y"
    )

    def setUp(self) -> None:
        from josepy.jwk import JWKRSA

        self.jwk256 = JWKRSA(key=RSA256_KEY.public_key())
        self.jwk256json = {
            "kty": "RSA",
            "e": "AQAB",
            "n": "m2Fylv-Uz7trgTW8EBHP3FQSMeZs2GNQ6VRo1sIVJEk",
        }
        self.jwk256_not_comparable = JWKRSA(key=RSA256_KEY.public_key()._wrapped)
        self.jwk512 = JWKRSA(key=RSA512_KEY.public_key())
        self.jwk512json = {
            "kty": "RSA",
            "e": "AQAB",
            "n": "rHVztFHtH92ucFJD_N_HW9AsdRsUuHUBBBDlHwNlRd3fp5"
            "80rv2-6QWE30cWgdmJS86ObRz6lUTor4R0T-3C5Q",
        }
        self.private = JWKRSA(key=RSA256_KEY)
        self.private_json_small = self.jwk256json.copy()
        self.private_json_small["d"] = "lPQED_EPTV0UIBfNI3KP2d9Jlrc2mrMllmf946bu-CE"
        self.private_json = self.jwk256json.copy()
        self.private_json.update(
            {
                "d": "lPQED_EPTV0UIBfNI3KP2d9Jlrc2mrMllmf946bu-CE",
                "p": "zUVNZn4lLLBD1R6NE8TKNQ",
                "q": "wcfKfc7kl5jfqXArCRSURQ",
                "dp": "CWJFq43QvT5Bm5iN8n1okQ",
                "dq": "bHh2u7etM8LKKCF2pY2UdQ",
                "qi": "oi45cEkbVoJjAbnQpFY87Q",
            }
        )
        self.jwk = self.private

    def test_init_auto_comparable(self) -> None:
        assert isinstance(self.jwk256_not_comparable.key, util.ComparableRSAKey)
        assert self.jwk256 == self.jwk256_not_comparable

    def test_encode_param_zero(self) -> None:
        from josepy.jwk import JWKRSA

        # TODO: move encode/decode _param to separate class
        assert "AA" == JWKRSA._encode_param(0)

    def test_equals(self) -> None:
        assert self.jwk256 == self.jwk256
        assert self.jwk512 == self.jwk512

    def test_not_equals(self) -> None:
        assert self.jwk256 != self.jwk512
        assert self.jwk512 != self.jwk256

    def test_load(self) -> None:
        from josepy.jwk import JWKRSA

        assert self.private == JWKRSA.load(test_util.load_vector("rsa256_key.pem"))

    def test_public_key(self) -> None:
        assert self.jwk256 == self.private.public_key()

    def test_to_partial_json(self) -> None:
        assert self.jwk256.to_partial_json() == self.jwk256json
        assert self.jwk512.to_partial_json() == self.jwk512json
        assert self.private.to_partial_json() == self.private_json

    def test_from_json(self) -> None:
        from josepy.jwk import JWK

        assert self.jwk256 == JWK.from_json(self.jwk256json)
        assert self.jwk512 == JWK.from_json(self.jwk512json)
        assert self.private == JWK.from_json(self.private_json)

    def test_from_json_private_small(self) -> None:
        from josepy.jwk import JWK

        assert self.private == JWK.from_json(self.private_json_small)

    def test_from_json_missing_one_additional(self) -> None:
        from josepy.jwk import JWK

        del self.private_json["q"]
        with pytest.raises(errors.Error):
            JWK.from_json(self.private_json)

    def test_from_json_hashable(self) -> None:
        from josepy.jwk import JWK

        hash(JWK.from_json(self.jwk256json))

    def test_from_json_non_schema_errors(self) -> None:
        # valid against schema, but still failing
        from josepy.jwk import JWK

        with pytest.raises(errors.DeserializationError):
            JWK.from_json({"kty": "RSA", "e": "AQAB", "n": ""})
        with pytest.raises(errors.DeserializationError):
            JWK.from_json({"kty": "RSA", "e": "AQAB", "n": "1"})

    def test_thumbprint_go_jose(self) -> None:
        # https://github.com/square/go-jose/blob/4ddd71883fa547d37fbf598071f04512d8bafee3/jwk.go#L155
        # https://github.com/square/go-jose/blob/4ddd71883fa547d37fbf598071f04512d8bafee3/jwk_test.go#L331-L344
        # https://github.com/square/go-jose/blob/4ddd71883fa547d37fbf598071f04512d8bafee3/jwk_test.go#L384
        from josepy.jwk import JWKRSA

        key = JWKRSA.json_loads(
            """{
    "kty": "RSA",
    "kid": "bilbo.baggins@hobbiton.example",
    "use": "sig",
    "n": "n4EPtAOCc9AlkeQHPzHStgAbgs7bTZLwUBZdR8_KuKPEHLd4rHVTeT-O-XV2jRojdNhxJWTDvNd7nqQ0VEiZQHz_AJmSCpMaJMRBSFKrKb2wqVwGU_NsYOYL-QtiWN2lbzcEe6XC0dApr5ydQLrHqkHHig3RBordaZ6Aj-oBHqFEHYpPe7Tpe-OfVfHd1E6cS6M1FZcD1NNLYD5lFHpPI9bTwJlsde3uhGqC0ZCuEHg8lhzwOHrtIQbS0FVbb9k3-tVTU4fg_3L_vniUFAKwuCLqKnS2BYwdq_mzSnbLY7h_qixoR7jig3__kRhuaxwUkRz5iaiQkqgc5gHdrNP5zw",
    "e": "AQAB"
}"""  # noqa
        )
        assert (
            binascii.hexlify(key.thumbprint())
            == b"f63838e96077ad1fc01c3f8405774dedc0641f558ebb4b40dccf5f9b6d66a932"
        )


class JWKECTest(unittest.TestCase, JWKTestBaseMixin):
    """Tests for josepy.jwk.JWKEC."""

    # pylint: disable=too-many-instance-attributes

    thumbprint = (
        b"\x06\xceL\x1b\xa8\x8d\x86\x1flF\x99J\x8b\xe0$\t\xbbj"
        b"\xd8\xf6O\x1ed\xdeR\x8f\x97\xff\xf6\xa2\x86\xd3"
    )

    def setUp(self) -> None:
        from josepy.jwk import JWKEC

        self.jwk256 = JWKEC(key=EC_P256_KEY.public_key())
        self.jwk384 = JWKEC(key=EC_P384_KEY.public_key())
        self.jwk521 = JWKEC(key=EC_P521_KEY.public_key())
        self.jwk256_not_comparable = JWKEC(key=EC_P256_KEY.public_key()._wrapped)
        self.jwk256json = {
            "kty": "EC",
            "crv": "P-256",
            "x": "jjQtV-fA7J_tK8dPzYq7jRPNjF8r5p6LW2R25S2Gw5U",
            "y": "EPAw8_8z7PYKsHH6hlGSlsWxFoFl7-0vM0QRGbmnvCc",
        }
        self.jwk384json = {
            "kty": "EC",
            "crv": "P-384",
            "x": "tIhpNtEXkadUbrY84rYGgApFM1X_3l3EWQRuOP1IWtxlTftrZQwneJZF0k0eRn00",
            "y": "KW2Gp-TThDXmZ-9MJPnD8hv-X130SVvfZRl1a04HPVwIbvLe87mvA_iuOa-myUyv",
        }
        self.jwk521json = {
            "kty": "EC",
            "crv": "P-521",
            "x": "AFkdl6cKzBmP18U8fffpP4IZN2eED45hDcwRPl5ZeClwHcLtnMBMuWYFFO_Nzm6DL2MhpN0zI2bcMLJd95aY2tPs",  # noqa
            "y": "AYvZq3wByjt7nQd8nYMqhFNCL3j_-U6GPWZet1hYBY_XZHrC4yIV0R4JnssRAY9eqc1EElpCc4hziis1jiV1iR4W",  # noqa
        }
        self.private = JWKEC(key=EC_P256_KEY)
        self.private_json = {
            "d": "xReNQBKqqTthG8oTmBdhp4EQYImSK1dVqfa2yyMn2rc",
            "x": "jjQtV-fA7J_tK8dPzYq7jRPNjF8r5p6LW2R25S2Gw5U",
            "y": "EPAw8_8z7PYKsHH6hlGSlsWxFoFl7-0vM0QRGbmnvCc",
            "crv": "P-256",
            "kty": "EC",
        }
        self.jwk = self.private

    def test_init_auto_comparable(self) -> None:
        assert isinstance(self.jwk256_not_comparable.key, util.ComparableECKey)
        assert self.jwk256 == self.jwk256_not_comparable

    def test_encode_param_zero(self) -> None:
        from josepy.jwk import JWKEC

        # pylint: disable=protected-access
        # TODO: move encode/decode _param to separate class
        assert "AA" == JWKEC._encode_param(0, 1)

    def test_equals(self) -> None:
        assert self.jwk256 == self.jwk256
        assert self.jwk384 == self.jwk384
        assert self.jwk521 == self.jwk521

    def test_not_equals(self) -> None:
        assert self.jwk256 != self.jwk384
        assert self.jwk256 != self.jwk521
        assert self.jwk384 != self.jwk256
        assert self.jwk384 != self.jwk521
        assert self.jwk521 != self.jwk256
        assert self.jwk521 != self.jwk384

    def test_load(self) -> None:
        from josepy.jwk import JWKEC

        assert self.private == JWKEC.load(test_util.load_vector("ec_p256_key.pem"))

    def test_public_key(self) -> None:
        assert self.jwk256 == self.private.public_key()

    def test_to_partial_json(self) -> None:
        assert self.jwk256.to_partial_json() == self.jwk256json
        assert self.jwk384.to_partial_json() == self.jwk384json
        assert self.jwk521.to_partial_json() == self.jwk521json
        assert self.private.to_partial_json() == self.private_json

    def test_from_json(self) -> None:
        from josepy.jwk import JWK

        assert self.jwk256 == JWK.from_json(self.jwk256json)
        assert self.jwk384 == JWK.from_json(self.jwk384json)
        assert self.jwk521 == JWK.from_json(self.jwk521json)
        assert self.private == JWK.from_json(self.private_json)

    def test_from_json_missing_x_coordinate(self) -> None:
        from josepy.jwk import JWK

        del self.private_json["x"]
        with pytest.raises(KeyError):
            JWK.from_json(self.private_json)

    def test_from_json_missing_y_coordinate(self) -> None:
        from josepy.jwk import JWK

        del self.private_json["y"]
        with pytest.raises(KeyError):
            JWK.from_json(self.private_json)

    def test_from_json_hashable(self) -> None:
        from josepy.jwk import JWK

        hash(JWK.from_json(self.jwk256json))

    def test_from_json_non_schema_errors(self) -> None:
        # valid against schema, but still failing
        from josepy.jwk import JWK

        with pytest.raises(errors.DeserializationError):
            JWK.from_json(
                {
                    "kty": "EC",
                    "crv": "P-256",
                    "x": "AQAB",
                    "y": "m2Fylv-Uz7trgTW8EBHP3FQSMeZs2GNQ6VRo1sIVJEk",
                }
            )
        with pytest.raises(errors.DeserializationError):
            JWK.from_json(
                {
                    "kty": "EC",
                    "crv": "P-256",
                    "x": "jjQtV-fA7J_tK8dPzYq7jRPNjF8r5p6LW2R25S2Gw5U",
                    "y": "1",
                }
            )

    def test_unknown_crv_name(self) -> None:
        from josepy.jwk import JWK

        with pytest.raises(errors.DeserializationError):
            JWK.from_json(
                {
                    "kty": "EC",
                    "crv": "P-255",
                    "x": "jjQtV-fA7J_tK8dPzYq7jRPNjF8r5p6LW2R25S2Gw5U",
                    "y": "EPAw8_8z7PYKsHH6hlGSlsWxFoFl7-0vM0QRGbmnvCc",
                }
            )

    def test_encode_y_leading_zero_p256(self) -> None:
        import josepy
        from josepy.jwk import JWK, JWKEC

        data = b"""-----BEGIN EC PRIVATE KEY-----
MHcCAQEEICZ7LCI99Na2KZ/Fq8JmJROakGJ5+J7rHiGSPoO36kOAoAoGCCqGSM49
AwEHoUQDQgAEGS5RvStca15z2FEanCM3juoX7tE/LB7iD44GWawGE40APAl/iZuH
31wQfst4glTZpxkpEI/MzNZHjiYnqrGeSw==
-----END EC PRIVATE KEY-----"""
        key = JWKEC.load(data)
        json = key.to_partial_json()
        y = josepy.json_util.decode_b64jose(json["y"])
        assert y[0] == 0
        assert len(y) == 32
        JWK.from_json(json)


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



import importlib
import re
import sys
import warnings

import pytest

import josepy


@pytest.mark.skipif(sys.version_info[:2] != (3, 7), reason="requires Python 3.7")
def test_warns() -> None:
    with pytest.warns(DeprecationWarning, match=re.escape(r"Python 3.7 support")):
        importlib.reload(josepy)


@pytest.mark.skipif(sys.version_info[:2] == (3, 7), reason="requires Python != 3.7")
def test_does_not_warn() -> None:
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        importlib.reload(josepy)


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



"""Tests for josepy.interfaces."""
import sys
import unittest
from typing import Any, Dict, List

import pytest


class JSONDeSerializableTest(unittest.TestCase):
    def setUp(self) -> None:
        from josepy.interfaces import JSONDeSerializable

        class Basic(JSONDeSerializable):
            def __init__(self, v: Any) -> None:
                self.v = v

            def to_partial_json(self) -> Any:
                return self.v

            @classmethod
            def from_json(cls, jobj: Any) -> "Basic":
                return cls(jobj)

        class Sequence(JSONDeSerializable):
            def __init__(self, x: Basic, y: Basic) -> None:
                self.x = x
                self.y = y

            def to_partial_json(self) -> List[Basic]:
                return [self.x, self.y]

            @classmethod
            def from_json(cls, jobj: List[Any]) -> "Sequence":
                return cls(Basic.from_json(jobj[0]), Basic.from_json(jobj[1]))

        class Mapping(JSONDeSerializable):
            def __init__(self, x: Any, y: Any) -> None:
                self.x = x
                self.y = y

            def to_partial_json(self) -> Dict[Basic, Basic]:
                return {self.x: self.y}

            @classmethod
            def from_json(cls, jobj: Any) -> "Mapping":
                return cls("dummy", "values")  # pragma: no cover

        self.basic1 = Basic("foo1")
        self.basic2 = Basic("foo2")
        self.seq = Sequence(self.basic1, self.basic2)
        self.mapping = Mapping(self.basic1, self.basic2)
        self.nested = Basic([[self.basic1]])
        self.tuple = Basic(("foo",))

        self.Basic = Basic
        self.Sequence = Sequence
        self.Mapping = Mapping

    def test_to_json_sequence(self) -> None:
        assert self.seq.to_json() == ["foo1", "foo2"]

    def test_to_json_mapping(self) -> None:
        assert self.mapping.to_json() == {"foo1": "foo2"}

    def test_to_json_other(self) -> None:
        mock_value = object()
        assert self.Basic(mock_value).to_json() is mock_value

    def test_to_json_nested(self) -> None:
        assert self.nested.to_json() == [["foo1"]]

    def test_to_json(self) -> None:
        assert self.tuple.to_json() == (("foo",))

    def test_from_json_not_implemented(self) -> None:
        from josepy.interfaces import JSONDeSerializable

        with pytest.raises(TypeError):
            JSONDeSerializable.from_json("xxx")

    def test_json_loads(self) -> None:
        seq = self.Sequence.json_loads('["foo1", "foo2"]')
        assert isinstance(seq, self.Sequence)
        assert isinstance(seq.x, self.Basic)
        assert isinstance(seq.y, self.Basic)
        assert seq.x.v == "foo1"
        assert seq.y.v == "foo2"

    def test_json_dumps(self) -> None:
        assert '["foo1", "foo2"]' == self.seq.json_dumps()

    def test_json_dumps_pretty(self) -> None:
        assert self.seq.json_dumps_pretty() == '[\n    "foo1",\n    "foo2"\n]'

    def test_json_dump_default(self) -> None:
        from josepy.interfaces import JSONDeSerializable

        assert "foo1" == JSONDeSerializable.json_dump_default(self.basic1)

        jobj = JSONDeSerializable.json_dump_default(self.seq)
        assert len(jobj) == 2
        assert jobj[0] is self.basic1
        assert jobj[1] is self.basic2

    def test_json_dump_default_type_error(self) -> None:
        from josepy.interfaces import JSONDeSerializable

        with pytest.raises(TypeError):
            # We're purposefully testing with the incorrect type here.
            JSONDeSerializable.json_dump_default(object())  # type: ignore


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



"""Tests for josepy.jwa."""
import sys
import unittest
from typing import Any
from unittest import mock

import pytest
import test_util

from josepy import errors

RSA256_KEY = test_util.load_rsa_private_key("rsa256_key.pem")
RSA512_KEY = test_util.load_rsa_private_key("rsa512_key.pem")
RSA1024_KEY = test_util.load_rsa_private_key("rsa1024_key.pem")
EC_P256_KEY = test_util.load_ec_private_key("ec_p256_key.pem")
EC_P384_KEY = test_util.load_ec_private_key("ec_p384_key.pem")
EC_P521_KEY = test_util.load_ec_private_key("ec_p521_key.pem")


class JWASignatureTest(unittest.TestCase):
    """Tests for josepy.jwa.JWASignature."""

    def setUp(self) -> None:
        from josepy.jwa import JWASignature

        class MockSig(JWASignature):
            def sign(self, key: Any, msg: bytes) -> bytes:
                raise NotImplementedError()  # pragma: no cover

            def verify(self, key: Any, msg: bytes, sig: bytes) -> bool:
                raise NotImplementedError()  # pragma: no cover

        self.Sig1 = MockSig("Sig1")
        self.Sig2 = MockSig("Sig2")

    def test_eq(self) -> None:
        assert self.Sig1 == self.Sig1

    def test_ne(self) -> None:
        assert self.Sig1 != self.Sig2

    def test_ne_other_type(self) -> None:
        assert self.Sig1 != 5

    def test_repr(self) -> None:
        assert "Sig1" == repr(self.Sig1)
        assert "Sig2" == repr(self.Sig2)

    def test_to_partial_json(self) -> None:
        assert self.Sig1.to_partial_json() == "Sig1"
        assert self.Sig2.to_partial_json() == "Sig2"

    def test_from_json(self) -> None:
        from josepy.jwa import RS256, JWASignature

        assert JWASignature.from_json("RS256") is RS256


class JWAHSTest(unittest.TestCase):
    def test_it(self) -> None:
        from josepy.jwa import HS256

        sig = (
            b"\xceR\xea\xcd\x94\xab\xcf\xfb\xe0\xacA.:\x1a'\x08i\xe2\xc4"
            b"\r\x85+\x0e\x85\xaeUZ\xd4\xb3\x97zO"
        )
        assert HS256.sign(b"some key", b"foo") == sig
        assert HS256.verify(b"some key", b"foo", sig) is True
        assert HS256.verify(b"some key", b"foo", sig + b"!") is False


class JWARSTest(unittest.TestCase):
    def test_sign_no_private_part(self) -> None:
        from josepy.jwa import RS256

        with pytest.raises(errors.Error):
            RS256.sign(RSA512_KEY.public_key(), b"foo")

    def test_sign_key_too_small(self) -> None:
        from josepy.jwa import PS256, RS256

        with pytest.raises(errors.Error):
            RS256.sign(RSA256_KEY, b"foo")
        with pytest.raises(errors.Error):
            PS256.sign(RSA256_KEY, b"foo")

    def test_rs(self) -> None:
        from josepy.jwa import RS256

        sig = (
            b"|\xc6\xb2\xa4\xab(\x87\x99\xfa*:\xea\xf8\xa0N&}\x9f\x0f\xc0O"
            b'\xc6t\xa3\xe6\xfa\xbb"\x15Y\x80Y\xe0\x81\xb8\x88)\xba\x0c\x9c'
            b"\xa4\x99\x1e\x19&\xd8\xc7\x99S\x97\xfc\x85\x0cOV\xe6\x07\x99"
            b"\xd2\xb9.>}\xfd"
        )
        assert RS256.sign(RSA512_KEY, b"foo") == sig
        assert RS256.verify(RSA512_KEY.public_key(), b"foo", sig) is True
        assert RS256.verify(RSA512_KEY.public_key(), b"foo", sig + b"!") is False

    def test_ps(self) -> None:
        from josepy.jwa import PS256

        sig = PS256.sign(RSA1024_KEY, b"foo")
        assert PS256.verify(RSA1024_KEY.public_key(), b"foo", sig) is True
        assert PS256.verify(RSA1024_KEY.public_key(), b"foo", sig + b"!") is False

    def test_sign_new_api(self) -> None:
        from josepy.jwa import RS256

        key = mock.MagicMock()
        RS256.sign(key, b"message")
        assert key.sign.called is True

    def test_verify_new_api(self) -> None:
        from josepy.jwa import RS256

        key = mock.MagicMock()
        RS256.verify(key, b"message", b"signature")
        assert key.verify.called is True


class JWAECTest(unittest.TestCase):
    def test_sign_no_private_part(self) -> None:
        from josepy.jwa import ES256

        with pytest.raises(errors.Error):
            ES256.sign(EC_P256_KEY.public_key(), b"foo")

    def test_es256_sign_and_verify(self) -> None:
        from josepy.jwa import ES256

        message = b"foo"
        signature = ES256.sign(EC_P256_KEY, message)
        assert ES256.verify(EC_P256_KEY.public_key(), message, signature) is True

    def test_es384_sign_and_verify(self) -> None:
        from josepy.jwa import ES384

        message = b"foo"
        signature = ES384.sign(EC_P384_KEY, message)
        assert ES384.verify(EC_P384_KEY.public_key(), message, signature) is True

    def test_verify_with_wrong_jwa(self) -> None:
        from josepy.jwa import ES256, ES384

        message = b"foo"
        signature = ES256.sign(EC_P256_KEY, message)
        assert ES384.verify(EC_P384_KEY.public_key(), message, signature) is False

    def test_verify_with_different_key(self) -> None:
        from cryptography.hazmat.backends import default_backend
        from cryptography.hazmat.primitives.asymmetric import ec

        from josepy.jwa import ES256

        message = b"foo"
        signature = ES256.sign(EC_P256_KEY, message)
        different_key = ec.generate_private_key(ec.SECP256R1(), default_backend())
        assert ES256.verify(different_key.public_key(), message, signature) is False

    def test_sign_new_api(self) -> None:
        from cryptography.hazmat.primitives.asymmetric.ec import SECP256R1

        from josepy.jwa import ES256

        key = mock.MagicMock(curve=SECP256R1())
        with mock.patch("josepy.jwa.decode_dss_signature") as decode_patch:
            decode_patch.return_value = (0, 0)
            ES256.sign(key, b"message")
        assert key.sign.called is True

    def test_verify_new_api(self) -> None:
        import math

        from cryptography.hazmat.primitives.asymmetric.ec import SECP256R1

        from josepy.jwa import ES256

        key = mock.MagicMock(key_size=256, curve=SECP256R1())
        ES256.verify(key, b"message", b"\x00" * math.ceil(key.key_size / 8) * 2)
        assert key.verify.called is True

    def test_signature_size(self) -> None:
        from josepy.jwa import ES512
        from josepy.jwk import JWK

        key = JWK.from_json(
            {
                "d": (
                    "Af9KP6DqLRbtit6NS_LRIaCP_-NdC5l5R2ugbILdfpv6dS9R4wUPNxiGw"
                    "-vVWumA56Yo1oBnEm8ZdR4W-u1lPHw5"
                ),
                "x": (
                    "AD4i4STyJ07iZJkHkpKEOuICpn6IHknzwAlrf-1w1a5dqOsRe30EECSN4vFxae"
                    "AmtdBSCKBwCq7h1q4bPgMrMUvF"
                ),
                "y": (
                    "AHAlXxrabjcx_yBxGObnm_DkEQMJK1E69OHY3x3VxF5VXoKc93CG4GLoaPvphZQv"
                    "Znt5EfExQoPktwOMIVhBHaFR"
                ),
                "crv": "P-521",
                "kty": "EC",
            }
        )
        with mock.patch("josepy.jwa.decode_dss_signature") as decode_patch:
            decode_patch.return_value = (0, 0)
            assert isinstance(key, JWK)
            sig = ES512.sign(key.key, b"test")
            assert len(sig) == 2 * 66


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



"""Tests for josepy.magic_typing."""
import sys
import warnings
from unittest import mock

import pytest


def test_import_success() -> None:
    import typing as temp_typing

    typing_class_mock = mock.MagicMock()
    text_mock = mock.MagicMock()
    typing_class_mock.Text = text_mock
    sys.modules["typing"] = typing_class_mock
    if "josepy.magic_typing" in sys.modules:
        del sys.modules["josepy.magic_typing"]  # pragma: no cover
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        from josepy.magic_typing import Text
    assert Text == text_mock
    del sys.modules["josepy.magic_typing"]
    sys.modules["typing"] = temp_typing


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



"""Tests for josepy.jws."""
import base64
import sys
import unittest
from unittest import mock

import OpenSSL
import pytest
import test_util

from josepy import errors, json_util, jwa, jwk

CERT = test_util.load_comparable_cert("cert.pem")
KEY = jwk.JWKRSA.load(test_util.load_vector("rsa512_key.pem"))


class MediaTypeTest(unittest.TestCase):
    """Tests for josepy.jws.MediaType."""

    def test_decode(self) -> None:
        from josepy.jws import MediaType

        assert "application/app" == MediaType.decode("application/app")
        assert "application/app" == MediaType.decode("app")
        with pytest.raises(errors.DeserializationError):
            MediaType.decode("app;foo")

    def test_encode(self) -> None:
        from josepy.jws import MediaType

        assert "app" == MediaType.encode("application/app")
        assert "application/app;foo" == MediaType.encode("application/app;foo")


class HeaderTest(unittest.TestCase):
    """Tests for josepy.jws.Header."""

    def setUp(self) -> None:
        from josepy.jws import Header

        self.header1 = Header(jwk="foo")
        self.header2 = Header(jwk="bar")
        self.crit = Header(crit=("a", "b"))
        self.empty = Header()

    def test_add_non_empty(self) -> None:
        from josepy.jws import Header

        assert Header(jwk="foo", crit=("a", "b")) == self.header1 + self.crit

    def test_add_empty(self) -> None:
        assert self.header1 == self.header1 + self.empty
        assert self.header1 == self.empty + self.header1

    def test_add_overlapping_error(self) -> None:
        with pytest.raises(TypeError):
            self.header1.__add__(self.header2)

    def test_add_wrong_type_error(self) -> None:
        with pytest.raises(TypeError):
            self.header1.__add__("xxx")

    def test_crit_decode_always_errors(self) -> None:
        from josepy.jws import Header

        with pytest.raises(errors.DeserializationError):
            Header.from_json({"crit": ["a", "b"]})

    def test_x5c_decoding(self) -> None:
        from josepy.jws import Header

        header = Header(x5c=(CERT, CERT))
        jobj = header.to_partial_json()
        assert isinstance(CERT.wrapped, OpenSSL.crypto.X509)
        cert_asn1 = OpenSSL.crypto.dump_certificate(OpenSSL.crypto.FILETYPE_ASN1, CERT.wrapped)
        cert_b64 = base64.b64encode(cert_asn1)
        assert jobj == {"x5c": [cert_b64, cert_b64]}
        assert header == Header.from_json(jobj)
        jobj["x5c"][0] = base64.b64encode(b"xxx" + cert_asn1)
        with pytest.raises(errors.DeserializationError):
            Header.from_json(jobj)

    def test_find_key(self) -> None:
        assert "foo" == self.header1.find_key()
        assert "bar" == self.header2.find_key()
        with pytest.raises(errors.Error):
            self.crit.find_key()


class SignatureTest(unittest.TestCase):
    """Tests for josepy.jws.Signature."""

    def test_from_json(self) -> None:
        from josepy.jws import Header, Signature

        assert Signature(signature=b"foo", header=Header(alg=jwa.RS256)) == Signature.from_json(
            {"signature": "Zm9v", "header": {"alg": "RS256"}}
        )

    def test_from_json_no_alg_error(self) -> None:
        from josepy.jws import Signature

        with pytest.raises(errors.DeserializationError):
            Signature.from_json({"signature": "foo"})


class JWSTest(unittest.TestCase):
    """Tests for josepy.jws.JWS."""

    def setUp(self) -> None:
        self.privkey = KEY
        self.pubkey = self.privkey.public_key()

        from josepy.jws import JWS

        self.unprotected = JWS.sign(payload=b"foo", key=self.privkey, alg=jwa.RS256)
        self.protected = JWS.sign(
            payload=b"foo", key=self.privkey, alg=jwa.RS256, protect=frozenset(["jwk", "alg"])
        )
        self.mixed = JWS.sign(
            payload=b"foo", key=self.privkey, alg=jwa.RS256, protect=frozenset(["alg"])
        )

    def test_pubkey_jwk(self) -> None:
        assert self.unprotected.signature.combined.jwk == self.pubkey
        assert self.protected.signature.combined.jwk == self.pubkey
        assert self.mixed.signature.combined.jwk == self.pubkey

    def test_sign_unprotected(self) -> None:
        assert self.unprotected.verify() is True

    def test_sign_protected(self) -> None:
        assert self.protected.verify() is True

    def test_sign_mixed(self) -> None:
        assert self.mixed.verify() is True

    def test_compact_lost_unprotected(self) -> None:
        compact = self.mixed.to_compact()
        assert (
            b"eyJhbGciOiAiUlMyNTYifQ.Zm9v.OHdxFVj73l5LpxbFp1AmYX4yJM0Pyb"
            b"_893n1zQjpim_eLS5J1F61lkvrCrCDErTEJnBGOGesJ72M7b6Ve1cAJA" == compact
        )

        from josepy.jws import JWS

        mixed = JWS.from_compact(compact)

        assert self.mixed != mixed
        assert {"alg"} == set(mixed.signature.combined.not_omitted())

    def test_from_compact_missing_components(self) -> None:
        from josepy.jws import JWS

        with pytest.raises(errors.DeserializationError):
            JWS.from_compact(b".")

    def test_json_omitempty(self) -> None:
        protected_jobj = self.protected.to_partial_json(flat=True)
        unprotected_jobj = self.unprotected.to_partial_json(flat=True)

        assert "protected" not in unprotected_jobj
        assert "header" not in protected_jobj

        unprotected_jobj["header"] = unprotected_jobj["header"].to_json()

        from josepy.jws import JWS

        assert JWS.from_json(protected_jobj) == self.protected
        assert JWS.from_json(unprotected_jobj) == self.unprotected

    def test_json_flat(self) -> None:
        jobj_to = {
            "signature": json_util.encode_b64jose(self.mixed.signature.signature),
            "payload": json_util.encode_b64jose(b"foo"),
            "header": self.mixed.signature.header,
            "protected": json_util.encode_b64jose(self.mixed.signature.protected.encode("utf-8")),
        }

        from josepy.jws import Header

        jobj_from = jobj_to.copy()
        header = jobj_from["header"]
        assert isinstance(header, Header)
        jobj_from["header"] = header.to_json()

        assert self.mixed.to_partial_json(flat=True) == jobj_to
        from josepy.jws import JWS

        assert self.mixed == JWS.from_json(jobj_from)

    def test_json_not_flat(self) -> None:
        jobj_to = {
            "signatures": (self.mixed.signature,),
            "payload": json_util.encode_b64jose(b"foo"),
        }
        jobj_from = jobj_to.copy()
        signature = jobj_to["signatures"][0]
        from josepy.jws import Signature

        assert isinstance(signature, Signature)
        jobj_from["signatures"] = [signature.to_json()]

        assert self.mixed.to_partial_json(flat=False) == jobj_to
        from josepy.jws import JWS

        assert self.mixed == JWS.from_json(jobj_from)

    def test_from_json_mixed_flat(self) -> None:
        from josepy.jws import JWS

        with pytest.raises(errors.DeserializationError):
            JWS.from_json({"signatures": (), "signature": "foo"})

    def test_from_json_hashable(self) -> None:
        from josepy.jws import JWS

        hash(JWS.from_json(self.mixed.to_json()))


class CLITest(unittest.TestCase):
    def setUp(self) -> None:
        self.key_path = test_util.vector_path("rsa512_key.pem")

    def test_unverified(self) -> None:
        from josepy.jws import CLI

        with mock.patch("sys.stdin") as sin:
            sin.read.return_value = '{"payload": "foo", "signature": "xxx"}'
            with mock.patch("sys.stdout"):
                assert CLI.run(["verify"]) is False

    def test_json(self) -> None:
        from josepy.jws import CLI

        with mock.patch("sys.stdin") as sin:
            sin.read.return_value = "foo"
            with mock.patch("sys.stdout") as sout:
                CLI.run(["sign", "-k", self.key_path, "-a", "RS256", "-p", "jwk"])
                sin.read.return_value = sout.write.mock_calls[0][1][0]
                assert 0 == CLI.run(["verify"])

    def test_compact(self) -> None:
        from josepy.jws import CLI

        with mock.patch("sys.stdin") as sin:
            sin.read.return_value = "foo"
            with mock.patch("sys.stdout") as sout:
                CLI.run(["--compact", "sign", "-k", self.key_path])
                sin.read.return_value = sout.write.mock_calls[0][1][0]
                assert 0 == CLI.run(["--compact", "verify", "--kty", "RSA", "-k", self.key_path])


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



"""Test utilities."""
import atexit
import contextlib
import os
import sys
from typing import Any

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from OpenSSL import crypto

import josepy.util
from josepy import ComparableRSAKey, ComparableX509
from josepy.util import ComparableECKey

# This approach is based on the recommendation at
# https://github.com/python/mypy/issues/1153#issuecomment-1207333806.
if sys.version_info >= (3, 9):
    import importlib.resources as importlib_resources
else:
    import importlib_resources

TESTDATA = importlib_resources.files("testdata")


def vector_path(*names: str) -> str:
    """Path to a test vector."""
    # This code is based on the recommendation at
    # https://web.archive.org/web/20230131043552/https://importlib-resources.readthedocs.io/en/latest/migration.html#pkg-resources-resource-filename.
    file_manager = contextlib.ExitStack()
    atexit.register(file_manager.close)
    ref = TESTDATA.joinpath(*names)
    # We convert the value to str here because some of the calling code doesn't
    # work with pathlib objects.
    return str(file_manager.enter_context(importlib_resources.as_file(ref)))


def load_vector(*names: str) -> bytes:
    """Load contents of a test vector."""
    return TESTDATA.joinpath(*names).read_bytes()


def _guess_loader(filename: str, loader_pem: Any, loader_der: Any) -> Any:
    _, ext = os.path.splitext(filename)
    if ext.lower() == ".pem":
        return loader_pem
    elif ext.lower() == ".der":
        return loader_der
    else:  # pragma: no cover
        raise ValueError("Loader could not be recognized based on extension")


def load_cert(*names: str) -> crypto.X509:
    """Load certificate."""
    loader = _guess_loader(names[-1], crypto.FILETYPE_PEM, crypto.FILETYPE_ASN1)
    return crypto.load_certificate(loader, load_vector(*names))


def load_comparable_cert(*names: str) -> josepy.util.ComparableX509:
    """Load ComparableX509 cert."""
    return ComparableX509(load_cert(*names))


def load_csr(*names: str) -> crypto.X509Req:
    """Load certificate request."""
    loader = _guess_loader(names[-1], crypto.FILETYPE_PEM, crypto.FILETYPE_ASN1)
    return crypto.load_certificate_request(loader, load_vector(*names))


def load_comparable_csr(*names: str) -> josepy.util.ComparableX509:
    """Load ComparableX509 certificate request."""
    return ComparableX509(load_csr(*names))


def load_rsa_private_key(*names: str) -> josepy.util.ComparableRSAKey:
    """Load RSA private key."""
    loader = _guess_loader(
        names[-1], serialization.load_pem_private_key, serialization.load_der_private_key
    )
    return ComparableRSAKey(loader(load_vector(*names), password=None, backend=default_backend()))


def load_ec_private_key(*names: str) -> josepy.util.ComparableECKey:
    """Load EC private key."""
    loader = _guess_loader(
        names[-1], serialization.load_pem_private_key, serialization.load_der_private_key
    )
    return ComparableECKey(loader(load_vector(*names), password=None, backend=default_backend()))


def load_pyopenssl_private_key(*names: str) -> crypto.PKey:
    """Load pyOpenSSL private key."""
    loader = _guess_loader(names[-1], crypto.FILETYPE_PEM, crypto.FILETYPE_ASN1)
    return crypto.load_privatekey(loader, load_vector(*names))



"""Tests for josepy.util."""
import functools
import sys
import unittest

import pytest
import test_util


class ComparableX509Test(unittest.TestCase):
    """Tests for josepy.util.ComparableX509."""

    def setUp(self) -> None:
        # test_util.load_comparable_{csr,cert} return ComparableX509
        self.req1 = test_util.load_comparable_csr("csr.pem")
        self.req2 = test_util.load_comparable_csr("csr.pem")
        self.req_other = test_util.load_comparable_csr("csr-san.pem")

        self.cert1 = test_util.load_comparable_cert("cert.pem")
        self.cert2 = test_util.load_comparable_cert("cert.pem")
        self.cert_other = test_util.load_comparable_cert("cert-san.pem")

    def test_getattr_proxy(self) -> None:
        assert self.cert1.has_expired() is True

    def test_eq(self) -> None:
        assert self.req1 == self.req2
        assert self.cert1 == self.cert2

    def test_ne(self) -> None:
        assert self.req1 != self.req_other
        assert self.cert1 != self.cert_other

    def test_ne_wrong_types(self) -> None:
        assert self.req1 != 5
        assert self.cert1 != 5

    def test_hash(self) -> None:
        assert hash(self.req1) == hash(self.req2)
        assert hash(self.req1) != hash(self.req_other)

        assert hash(self.cert1) == hash(self.cert2)
        assert hash(self.cert1) != hash(self.cert_other)

    def test_repr(self) -> None:
        for x509 in self.req1, self.cert1:
            assert repr(x509) == "<ComparableX509({0!r})>".format(x509.wrapped)


class ComparableRSAKeyTest(unittest.TestCase):
    """Tests for josepy.util.ComparableRSAKey."""

    def setUp(self) -> None:
        # test_utl.load_rsa_private_key return ComparableRSAKey
        self.key = test_util.load_rsa_private_key("rsa256_key.pem")
        self.key_same = test_util.load_rsa_private_key("rsa256_key.pem")
        self.key2 = test_util.load_rsa_private_key("rsa512_key.pem")

    def test_getattr_proxy(self) -> None:
        assert 256 == self.key.key_size

    def test_eq(self) -> None:
        assert self.key == self.key_same

    def test_ne(self) -> None:
        assert self.key != self.key2

    def test_ne_different_types(self) -> None:
        assert self.key != 5

    def test_ne_not_wrapped(self) -> None:
        assert self.key != self.key_same._wrapped

    def test_ne_no_serialization(self) -> None:
        from josepy.util import ComparableRSAKey

        assert ComparableRSAKey(5) != ComparableRSAKey(5)  # type: ignore

    def test_hash(self) -> None:
        assert isinstance(hash(self.key), int)
        assert hash(self.key) == hash(self.key_same)
        assert hash(self.key) != hash(self.key2)

    def test_repr(self) -> None:
        assert repr(self.key).startswith("<ComparableRSAKey(<cryptography.hazmat.") is True

    def test_public_key(self) -> None:
        from josepy.util import ComparableRSAKey

        assert isinstance(self.key.public_key(), ComparableRSAKey)


class ComparableECKeyTest(unittest.TestCase):
    """Tests for josepy.util.ComparableECKey."""

    def setUp(self) -> None:
        # test_utl.load_ec_private_key return ComparableECKey
        self.p256_key = test_util.load_ec_private_key("ec_p256_key.pem")
        self.p256_key_same = test_util.load_ec_private_key("ec_p256_key.pem")
        self.p384_key = test_util.load_ec_private_key("ec_p384_key.pem")
        self.p521_key = test_util.load_ec_private_key("ec_p521_key.pem")

    def test_getattr_proxy(self) -> None:
        assert 256 == self.p256_key.key_size

    def test_eq(self) -> None:
        assert self.p256_key == self.p256_key_same

    def test_ne(self) -> None:
        assert self.p256_key != self.p384_key
        assert self.p256_key != self.p521_key

    def test_ne_different_types(self) -> None:
        assert self.p256_key != 5

    def test_ne_not_wrapped(self) -> None:
        assert self.p256_key != self.p256_key_same._wrapped

    def test_ne_no_serialization(self) -> None:
        from josepy.util import ComparableECKey

        assert ComparableECKey(5) != ComparableECKey(5)  # type: ignore

    def test_hash(self) -> None:
        assert isinstance(hash(self.p256_key), int)
        assert hash(self.p256_key) == hash(self.p256_key_same)
        assert hash(self.p256_key) != hash(self.p384_key)
        assert hash(self.p256_key) != hash(self.p521_key)

    def test_repr(self) -> None:
        assert repr(self.p256_key).startswith("<ComparableECKey(<cryptography.hazmat.") is True

    def test_public_key(self) -> None:
        from josepy.util import ComparableECKey

        assert isinstance(self.p256_key.public_key(), ComparableECKey)


class ImmutableMapTest(unittest.TestCase):
    """Tests for josepy.util.ImmutableMap."""

    def setUp(self) -> None:
        from josepy.util import ImmutableMap

        class A(ImmutableMap):
            x: int
            y: int
            __slots__ = ("x", "y")

        class B(ImmutableMap):
            x: int
            y: int
            __slots__ = ("x", "y")

        self.A = A
        self.B = B

        self.a1 = self.A(x=1, y=2)
        self.a1_swap = self.A(y=2, x=1)
        self.a2 = self.A(x=3, y=4)
        self.b = self.B(x=1, y=2)

    def test_update(self) -> None:
        assert self.A(x=2, y=2) == self.a1.update(x=2)
        assert self.a2 == self.a1.update(x=3, y=4)

    def test_get_missing_item_raises_key_error(self) -> None:
        with pytest.raises(KeyError):
            self.a1.__getitem__("z")

    def test_order_of_args_does_not_matter(self) -> None:
        assert self.a1 == self.a1_swap

    def test_type_error_on_missing(self) -> None:
        with pytest.raises(TypeError):
            self.A(x=1)
        with pytest.raises(TypeError):
            self.A(y=2)

    def test_type_error_on_unrecognized(self) -> None:
        with pytest.raises(TypeError):
            self.A(x=1, z=2)
        with pytest.raises(TypeError):
            self.A(x=1, y=2, z=3)

    def test_get_attr(self) -> None:
        assert 1 == self.a1.x
        assert 2 == self.a1.y
        assert 1 == self.a1_swap.x
        assert 2 == self.a1_swap.y

    def test_set_attr_raises_attribute_error(self) -> None:
        with pytest.raises(AttributeError):
            functools.partial(self.a1.__setattr__, "x")(10)

    def test_equal(self) -> None:
        assert self.a1 == self.a1
        assert self.a2 == self.a2
        assert self.a1 != self.a2

    def test_hash(self) -> None:
        assert hash((1, 2)) == hash(self.a1)

    def test_unhashable(self) -> None:
        with pytest.raises(TypeError):
            self.A(x=1, y={}).__hash__()

    def test_repr(self) -> None:
        assert "A(x=1, y=2)" == repr(self.a1)
        assert "A(x=1, y=2)" == repr(self.a1_swap)
        assert "B(x=1, y=2)" == repr(self.b)
        assert "B(x='foo', y='bar')" == repr(self.B(x="foo", y="bar"))


class frozendictTest(unittest.TestCase):
    """Tests for josepy.util.frozendict."""

    def setUp(self) -> None:
        from josepy.util import frozendict

        self.fdict = frozendict(x=1, y="2")

    def test_init_dict(self) -> None:
        from josepy.util import frozendict

        assert self.fdict == frozendict({"x": 1, "y": "2"})

    def test_init_other_raises_type_error(self) -> None:
        from josepy.util import frozendict

        # specifically fail for generators...
        with pytest.raises(TypeError):
            frozendict({"a": "b"}.items())

    def test_len(self) -> None:
        assert 2 == len(self.fdict)

    def test_hash(self) -> None:
        assert isinstance(hash(self.fdict), int)

    def test_getattr_proxy(self) -> None:
        assert 1 == self.fdict.x
        assert "2" == self.fdict.y

    def test_getattr_raises_attribute_error(self) -> None:
        with pytest.raises(AttributeError):
            self.fdict.__getattr__("z")

    def test_setattr_immutable(self) -> None:
        with pytest.raises(AttributeError):
            self.fdict.__setattr__("z", 3)

    def test_repr(self) -> None:
        assert "frozendict(x=1, y='2')" == repr(self.fdict)


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover



"""Tests for josepy.json_util."""
import itertools
import sys
import unittest
from typing import Any, Dict, Mapping
from unittest import mock

import pytest
import test_util

from josepy import errors, interfaces, util

CERT = test_util.load_comparable_cert("cert.pem")
CSR = test_util.load_comparable_csr("csr.pem")


class FieldTest(unittest.TestCase):
    """Tests for josepy.json_util.field and josepy.json_util.Field."""

    def test_field_function(self) -> None:
        from josepy.json_util import Field, field

        test = field("foo", default="bar")
        assert isinstance(test, Field)
        assert test.json_name == "foo"
        assert test.default == "bar"

    def test_type_field_control(self) -> None:
        from josepy.json_util import JSONObjectWithFields, field

        class DummyProperlyTyped(JSONObjectWithFields):
            type: str = field("type")
            index: int = field("index")

        with pytest.raises(ValueError):

            class DummyImproperlyTyped(JSONObjectWithFields):
                type = field("type")
                index: int = field("index")

    def test_no_omit_boolean(self) -> None:
        from josepy.json_util import Field

        for default, omitempty, value in itertools.product(
            [True, False], [True, False], [True, False]
        ):
            assert Field("foo", default=default, omitempty=omitempty).omit(value) is False

    def test_descriptors(self) -> None:
        mock_value = mock.MagicMock()

        def decoder(unused_value: Any) -> str:
            return "d"

        def encoder(unused_value: Any) -> str:
            return "e"

        from josepy.json_util import Field

        field = Field("foo")

        field = field.encoder(encoder)
        assert "e" == field.encode(mock_value)

        field = field.decoder(decoder)
        assert "e" == field.encode(mock_value)
        assert "d" == field.decode(mock_value)

    def test_default_encoder_is_partial(self) -> None:
        class MockField(interfaces.JSONDeSerializable):
            def to_partial_json(self) -> Dict[str, Any]:
                return {"foo": "bar"}  # pragma: no cover

            @classmethod
            def from_json(cls, jobj: Mapping[str, Any]) -> "MockField":
                return cls()  # pragma: no cover

        mock_field = MockField()

        from josepy.json_util import Field

        assert Field.default_encoder(mock_field) is mock_field
        # in particular...
        assert "foo" != Field.default_encoder(mock_field)

    def test_default_encoder_passthrough(self) -> None:
        mock_value = mock.MagicMock()
        from josepy.json_util import Field

        assert Field.default_encoder(mock_value) is mock_value

    def test_default_decoder_list_to_tuple(self) -> None:
        from josepy.json_util import Field

        assert (1, 2, 3) == Field.default_decoder([1, 2, 3])

    def test_default_decoder_dict_to_frozendict(self) -> None:
        from josepy.json_util import Field

        obj = Field.default_decoder({"x": 2})
        assert isinstance(obj, util.frozendict)
        assert obj == util.frozendict(x=2)

    def test_default_decoder_passthrough(self) -> None:
        mock_value = mock.MagicMock()
        from josepy.json_util import Field

        assert Field.default_decoder(mock_value) is mock_value


class JSONObjectWithFieldsMetaTest(unittest.TestCase):
    """Tests for josepy.json_util.JSONObjectWithFieldsMeta."""

    def setUp(self) -> None:
        from josepy.json_util import Field, JSONObjectWithFieldsMeta

        self.field = Field("Baz")
        self.field2 = Field("Baz2")

        class A(metaclass=JSONObjectWithFieldsMeta):
            __slots__ = ("bar",)
            baz = self.field

        class B(A):
            pass

        class C(A):
            baz = self.field2

        self.a_cls = A
        self.b_cls = B
        self.c_cls = C

    def test_fields(self) -> None:
        assert {"baz": self.field} == self.a_cls._fields
        assert {"baz": self.field} == self.b_cls._fields

    def test_fields_inheritance(self) -> None:
        assert {"baz": self.field2} == self.c_cls._fields

    def test_slots(self) -> None:
        assert ("bar", "baz") == self.a_cls.__slots__
        assert ("baz",) == self.b_cls.__slots__

    def test_orig_slots(self) -> None:
        assert ("bar",) == self.a_cls._orig_slots
        assert () == self.b_cls._orig_slots


class JSONObjectWithFieldsTest(unittest.TestCase):
    """Tests for josepy.json_util.JSONObjectWithFields."""

    def setUp(self) -> None:
        from josepy.json_util import Field, JSONObjectWithFields

        class MockJSONObjectWithFields(JSONObjectWithFields):
            x = Field("x", omitempty=True, encoder=(lambda x: x * 2), decoder=(lambda x: x / 2))
            y = Field("y")
            z = Field("Z")  # on purpose uppercase

            @y.encoder  # type: ignore
            def y(value):
                if value == 500:
                    raise errors.SerializationError()
                return value

            @y.decoder  # type: ignore
            def y(value):
                if value == 500:
                    raise errors.DeserializationError()
                return value

        self.MockJSONObjectWithFields = MockJSONObjectWithFields
        self.mock = MockJSONObjectWithFields(x=None, y=2, z=3)

    def test_init_defaults(self) -> None:
        assert self.mock == self.MockJSONObjectWithFields(y=2, z=3)

    def test_encode(self) -> None:
        assert 10 == self.MockJSONObjectWithFields(x=5, y=0, z=0).encode("x")

    def test_encode_wrong_field(self) -> None:
        with pytest.raises(errors.Error):
            self.mock.encode("foo")

    def test_encode_serialization_error_passthrough(self) -> None:
        with pytest.raises(errors.SerializationError):
            self.MockJSONObjectWithFields(y=500, z=None).encode("y")

    def test_fields_to_partial_json_omits_empty(self) -> None:
        assert self.mock.fields_to_partial_json() == {"y": 2, "Z": 3}

    def test_fields_from_json_fills_default_for_empty(self) -> None:
        assert {"x": None, "y": 2, "z": 3} == self.MockJSONObjectWithFields.fields_from_json(
            {"y": 2, "Z": 3}
        )

    def test_fields_from_json_fails_on_missing(self) -> None:
        with pytest.raises(errors.DeserializationError):
            self.MockJSONObjectWithFields.fields_from_json({"y": 0})
        with pytest.raises(errors.DeserializationError):
            self.MockJSONObjectWithFields.fields_from_json({"Z": 0})
        with pytest.raises(errors.DeserializationError):
            self.MockJSONObjectWithFields.fields_from_json({"x": 0, "y": 0})
        with pytest.raises(errors.DeserializationError):
            self.MockJSONObjectWithFields.fields_from_json({"x": 0, "Z": 0})

    def test_fields_to_partial_json_encoder(self) -> None:
        assert self.MockJSONObjectWithFields(x=1, y=2, z=3).to_partial_json() == {
            "x": 2,
            "y": 2,
            "Z": 3,
        }

    def test_fields_from_json_decoder(self) -> None:
        assert {"x": 2, "y": 2, "z": 3} == self.MockJSONObjectWithFields.fields_from_json(
            {"x": 4, "y": 2, "Z": 3}
        )

    def test_fields_to_partial_json_error_passthrough(self) -> None:
        with pytest.raises(errors.SerializationError):
            self.MockJSONObjectWithFields(x=1, y=500, z=3).to_partial_json()

    def test_fields_from_json_error_passthrough(self) -> None:
        with pytest.raises(errors.DeserializationError):
            self.MockJSONObjectWithFields.from_json({"x": 4, "y": 500, "Z": 3})


class DeEncodersTest(unittest.TestCase):
    def setUp(self) -> None:
        self.b64_cert = (
            "MIIB3jCCAYigAwIBAgICBTkwDQYJKoZIhvcNAQELBQAwdzELMAkGA1UEBhM"
            "CVVMxETAPBgNVBAgMCE1pY2hpZ2FuMRIwEAYDVQQHDAlBbm4gQXJib3IxKz"
            "ApBgNVBAoMIlVuaXZlcnNpdHkgb2YgTWljaGlnYW4gYW5kIHRoZSBFRkYxF"
            "DASBgNVBAMMC2V4YW1wbGUuY29tMB4XDTE0MTIxMTIyMzQ0NVoXDTE0MTIx"
            "ODIyMzQ0NVowdzELMAkGA1UEBhMCVVMxETAPBgNVBAgMCE1pY2hpZ2FuMRI"
            "wEAYDVQQHDAlBbm4gQXJib3IxKzApBgNVBAoMIlVuaXZlcnNpdHkgb2YgTW"
            "ljaGlnYW4gYW5kIHRoZSBFRkYxFDASBgNVBAMMC2V4YW1wbGUuY29tMFwwD"
            "QYJKoZIhvcNAQEBBQADSwAwSAJBAKx1c7RR7R_drnBSQ_zfx1vQLHUbFLh1"
            "AQQQ5R8DZUXd36efNK79vukFhN9HFoHZiUvOjm0c-pVE6K-EdE_twuUCAwE"
            "AATANBgkqhkiG9w0BAQsFAANBAC24z0IdwIVKSlntksllvr6zJepBH5fMnd"
            "fk3XJp10jT6VE-14KNtjh02a56GoraAvJAT5_H67E8GvJ_ocNnB_o"
        )
        self.b64_csr = (
            "MIIBXTCCAQcCAQAweTELMAkGA1UEBhMCVVMxETAPBgNVBAgMCE1pY2hpZ2F"
            "uMRIwEAYDVQQHDAlBbm4gQXJib3IxDDAKBgNVBAoMA0VGRjEfMB0GA1UECw"
            "wWVW5pdmVyc2l0eSBvZiBNaWNoaWdhbjEUMBIGA1UEAwwLZXhhbXBsZS5jb"
            "20wXDANBgkqhkiG9w0BAQEFAANLADBIAkEArHVztFHtH92ucFJD_N_HW9As"
            "dRsUuHUBBBDlHwNlRd3fp580rv2-6QWE30cWgdmJS86ObRz6lUTor4R0T-3"
            "C5QIDAQABoCkwJwYJKoZIhvcNAQkOMRowGDAWBgNVHREEDzANggtleGFtcG"
            "xlLmNvbTANBgkqhkiG9w0BAQsFAANBAHJH_O6BtC9aGzEVCMGOZ7z9iIRHW"
            "Szr9x_bOzn7hLwsbXPAgO1QxEwL-X-4g20Gn9XBE1N9W6HCIEut2d8wACg"
        )

    def test_encode_b64jose(self) -> None:
        from josepy.json_util import encode_b64jose

        encoded = encode_b64jose(b"x")
        assert isinstance(encoded, str)
        assert "eA" == encoded

    def test_decode_b64jose(self) -> None:
        from josepy.json_util import decode_b64jose

        decoded = decode_b64jose("eA")
        assert isinstance(decoded, bytes)
        assert b"x" == decoded

    def test_decode_b64jose_padding_error(self) -> None:
        from josepy.json_util import decode_b64jose

        with pytest.raises(errors.DeserializationError):
            decode_b64jose("x")

    def test_decode_b64jose_size(self) -> None:
        from josepy.json_util import decode_b64jose

        assert b"foo" == decode_b64jose("Zm9v", size=3)
        with pytest.raises(errors.DeserializationError):
            decode_b64jose("Zm9v", size=2)
        with pytest.raises(errors.DeserializationError):
            decode_b64jose("Zm9v", size=4)

    def test_decode_b64jose_minimum_size(self) -> None:
        from josepy.json_util import decode_b64jose

        assert b"foo" == decode_b64jose("Zm9v", size=3, minimum=True)
        assert b"foo" == decode_b64jose("Zm9v", size=2, minimum=True)
        with pytest.raises(errors.DeserializationError):
            decode_b64jose("Zm9v", size=4, minimum=True)

    def test_encode_hex16(self) -> None:
        from josepy.json_util import encode_hex16

        encoded = encode_hex16(b"foo")
        assert "666f6f" == encoded
        assert isinstance(encoded, str)

    def test_decode_hex16(self) -> None:
        from josepy.json_util import decode_hex16

        decoded = decode_hex16("666f6f")
        assert b"foo" == decoded
        assert isinstance(decoded, bytes)

    def test_decode_hex16_minimum_size(self) -> None:
        from josepy.json_util import decode_hex16

        assert b"foo" == decode_hex16("666f6f", size=3, minimum=True)
        assert b"foo" == decode_hex16("666f6f", size=2, minimum=True)
        with pytest.raises(errors.DeserializationError):
            decode_hex16("666f6f", size=4, minimum=True)

    def test_decode_hex16_odd_length(self) -> None:
        from josepy.json_util import decode_hex16

        with pytest.raises(errors.DeserializationError):
            decode_hex16("x")

    def test_encode_cert(self) -> None:
        from josepy.json_util import encode_cert

        assert self.b64_cert == encode_cert(CERT)

    def test_decode_cert(self) -> None:
        from josepy.json_util import decode_cert

        cert = decode_cert(self.b64_cert)
        assert isinstance(cert, util.ComparableX509)
        assert cert == CERT
        with pytest.raises(errors.DeserializationError):
            decode_cert("")

    def test_encode_csr(self) -> None:
        from josepy.json_util import encode_csr

        assert self.b64_csr == encode_csr(CSR)

    def test_decode_csr(self) -> None:
        from josepy.json_util import decode_csr

        csr = decode_csr(self.b64_csr)
        assert isinstance(csr, util.ComparableX509)
        assert csr == CSR
        with pytest.raises(errors.DeserializationError):
            decode_csr("")


class TypedJSONObjectWithFieldsTest(unittest.TestCase):
    def setUp(self) -> None:
        from josepy.json_util import TypedJSONObjectWithFields

        class MockParentTypedJSONObjectWithFields(TypedJSONObjectWithFields):
            TYPES = {}
            type_field_name = "type"

        @MockParentTypedJSONObjectWithFields.register
        class MockTypedJSONObjectWithFields(MockParentTypedJSONObjectWithFields):
            foo: str
            typ = "test"
            __slots__ = ("foo",)

            @classmethod
            def fields_from_json(cls, jobj: Mapping[str, Any]) -> Dict[str, Any]:
                return {"foo": jobj["foo"]}

            def fields_to_partial_json(self) -> Any:
                return {"foo": self.foo}

        self.parent_cls = MockParentTypedJSONObjectWithFields
        self.msg = MockTypedJSONObjectWithFields(foo="bar")

    def test_to_partial_json(self) -> None:
        assert self.msg.to_partial_json() == {
            "type": "test",
            "foo": "bar",
        }

    def test_from_json_non_dict_fails(self) -> None:
        for value in [[], (), 5, "asd"]:  # all possible input types
            with pytest.raises(errors.DeserializationError):
                # We're purposefully testing with the incorrect type here.
                self.parent_cls.from_json(value)  # type: ignore

    def test_from_json_dict_no_type_fails(self) -> None:
        with pytest.raises(errors.DeserializationError):
            self.parent_cls.from_json({})

    def test_from_json_unknown_type_fails(self) -> None:
        with pytest.raises(errors.UnrecognizedTypeError):
            self.parent_cls.from_json({"type": "bar"})

    def test_from_json_returns_obj(self) -> None:
        assert {"foo": "bar"} == self.parent_cls.from_json({"type": "test", "foo": "bar"})


if __name__ == "__main__":
    sys.exit(pytest.main(sys.argv[1:] + [__file__]))  # pragma: no cover
